"""General purpose iterator for all ImagingExtractor data."""

import math
from typing import Optional

import numpy as np
from roiextractors import ImagingExtractor
from tqdm import tqdm

from neuroconv.tools.hdmf import GenericDataChunkIterator


class ImagingExtractorDataChunkIterator(GenericDataChunkIterator):
    """DataChunkIterator for ImagingExtractor objects primarily used when writing imaging data to an NWB file."""

    def __init__(
        self,
        imaging_extractor: ImagingExtractor,
        buffer_gb: Optional[float] = None,
        buffer_shape: Optional[tuple] = None,
        chunk_mb: Optional[float] = None,
        chunk_shape: Optional[tuple] = None,
        display_progress: bool = False,
        progress_bar_class: Optional[tqdm] = None,
        progress_bar_options: Optional[dict] = None,
    ):
        """
        Initialize an Iterable object which returns DataChunks with data and their selections on each iteration.

        Parameters
        ----------
        imaging_extractor : ImagingExtractor
            The ImagingExtractor object which handles the data access.
        buffer_gb : float, optional
            The upper bound on size in gigabytes (GB) of each selection from the iteration.
            The buffer_shape will be set implicitly by this argument.
            Cannot be set if `buffer_shape` is also specified.
            The default is 1GB.
        buffer_shape : tuple, optional
            Manual specification of buffer shape to return on each iteration.
            Must be a multiple of chunk_shape along each axis.
            Cannot be set if `buffer_gb` is also specified.
            The default is None.
        chunk_mb : float, optional
            The upper bound on size in megabytes (MB) of the internal chunk for the HDF5 dataset.
            The chunk_shape will be set implicitly by this argument.
            Cannot be set if `chunk_shape` is also specified.
            The default is 10MB, as recommended by the HDF5 group.
            For more details, search the hdf5 documentation for "Improving IO Performance Compressed Datasets".
        chunk_shape : tuple, optional
            Manual specification of the internal chunk shape for the HDF5 dataset.
            Cannot be set if `chunk_mb` is also specified.
            The default is None.
        display_progress : bool, default=False
            Display a progress bar with iteration rate and estimated completion time.
        progress_bar_class : dict, optional
            The progress bar class to use.
            Defaults to tqdm.tqdm if the TQDM package is installed.
        progress_bar_options : dict, optional
            Dictionary of keyword arguments to be passed directly to tqdm.
            See https://github.com/tqdm/tqdm#parameters for options.
        """
        self.imaging_extractor = imaging_extractor

        assert not (buffer_gb and buffer_shape), "Only one of 'buffer_gb' or 'buffer_shape' can be specified!"
        assert not (chunk_mb and chunk_shape), "Only one of 'chunk_mb' or 'chunk_shape' can be specified!"

        if chunk_mb and buffer_gb:
            assert chunk_mb * 1e6 <= buffer_gb * 1e9, "chunk_mb must be less than or equal to buffer_gb!"

        if chunk_mb is None and chunk_shape is None:
            chunk_mb = 10.0

        self._maxshape = self._get_maxshape()
        self._dtype = self._get_dtype()
        if chunk_shape is None:
            chunk_shape = self._get_default_chunk_shape(chunk_mb=chunk_mb)

        if buffer_gb is None and buffer_shape is None:
            buffer_gb = 1.0

        if buffer_shape is None:
            buffer_shape = self._get_scaled_buffer_shape(buffer_gb=buffer_gb, chunk_shape=chunk_shape)

        super().__init__(
            buffer_shape=buffer_shape,
            chunk_shape=chunk_shape,
            display_progress=display_progress,
            progress_bar_class=progress_bar_class,
            progress_bar_options=progress_bar_options,
        )

    def _get_default_chunk_shape(self, chunk_mb: float) -> tuple:
        """Select the chunk_shape less than the threshold of chunk_mb while keeping the original image size."""
        assert chunk_mb > 0, f"chunk_mb ({chunk_mb}) must be greater than zero!"

        num_frames = self._maxshape[0]
        width = self._maxshape[1]
        height = self._maxshape[2]

        frame_size_bytes = width * height * self._dtype.itemsize
        chunk_size_bytes = chunk_mb * 1e6
        num_frames_per_chunk = int(chunk_size_bytes / frame_size_bytes)

        if len(self._maxshape) == 3:
            chunk_shape = (max(min(num_frames_per_chunk, num_frames), 1), width, height)
        elif len(self._maxshape) == 4:
            chunk_shape = (max(min(num_frames_per_chunk, num_frames), 1), width, height, 1)

        return chunk_shape

    def _get_scaled_buffer_shape(self, buffer_gb: float, chunk_shape: tuple) -> tuple:
        """Select the buffer_shape less than the threshold of buffer_gb that is also a multiple of the chunk_shape."""
        assert buffer_gb > 0, f"buffer_gb ({buffer_gb}) must be greater than zero!"
        assert all(np.array(chunk_shape) > 0), f"Some dimensions of chunk_shape ({chunk_shape}) are less than zero!"

        image_size = self._get_maxshape()[1:]
        min_buffer_shape = tuple([chunk_shape[0]]) + image_size
        scaling_factor = math.floor((buffer_gb * 1e9 / (math.prod(min_buffer_shape) * self._get_dtype().itemsize)))
        max_buffer_shape = tuple([int(scaling_factor * min_buffer_shape[0])]) + image_size
        scaled_buffer_shape = tuple(
            [
                min(max(int(dimension_length), chunk_shape[dimension_index]), self._get_maxshape()[dimension_index])
                for dimension_index, dimension_length in enumerate(max_buffer_shape)
            ]
        )
        return scaled_buffer_shape

    def _get_dtype(self) -> np.dtype:
        return self.imaging_extractor.get_dtype()

    def _get_maxshape(self) -> tuple:
        video_shape = (self.imaging_extractor.get_num_frames(),)
        image_shape = self.imaging_extractor.get_image_size()
        width, height = image_shape[1], image_shape[0]  # ROIExtractors convention is flipped
        video_shape += (width, height)
        if len(image_shape) == 3:
            depth = image_shape[2]
            video_shape += (depth,)
        return video_shape

    def _get_data(self, selection: tuple[slice]) -> np.ndarray:
        data = self.imaging_extractor.get_video(
            start_frame=selection[0].start,
            end_frame=selection[0].stop,
        )
        tranpose_axes = (0, 2, 1) if len(data.shape) == 3 else (0, 2, 1, 3)
        return data.transpose(tranpose_axes)[(slice(0, self.buffer_shape[0]),) + selection[1:]]



from ._submit_aws_batch_job import submit_aws_batch_job
from ._rclone_transfer_batch_job import rclone_transfer_batch_job
from ._deploy_neuroconv_batch_job import deploy_neuroconv_batch_job

__all__ = [
    "submit_aws_batch_job",
    "rclone_transfer_batch_job",
    "deploy_neuroconv_batch_job",
]



import struct
import wave
from pathlib import Path

import numpy as np


def create_24bit_wav_file(output_path, duration=1.0, sample_rate=44100, frequency=440):
    """
    Create a 24-bit (3-byte) WAV file with a sine wave.

    Parameters
    ----------
    output_path : str or Path
        Path to save the WAV file
    duration : float, default: 1.0
        Duration of the audio in seconds
    sample_rate : int, default: 44100
        Sample rate in Hz
    frequency : int, default: 440
        Frequency of the sine wave in Hz

    Returns
    -------
    str
        Path to the created WAV file
    """
    output_path = Path(output_path)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    # Generate a sine wave
    t = np.linspace(0, duration, int(sample_rate * duration), endpoint=False)
    audio_data = np.sin(2 * np.pi * frequency * t)

    # Scale to 24-bit range (-2^23 to 2^23-1)
    max_value = 2**23 - 1
    audio_data = (audio_data * max_value).astype(np.int32)

    # Create WAV file
    with wave.open(str(output_path), "wb") as wav_file:
        wav_file.setnchannels(1)  # Mono
        wav_file.setsampwidth(3)  # 3 bytes per sample (24-bit)
        wav_file.setframerate(sample_rate)

        # Pack the data manually as 24-bit (3-byte) values
        for sample in audio_data:
            # Convert to bytes, little-endian, 3 bytes
            packed_value = struct.pack("<i", sample)[:3]  # Take only the first 3 bytes
            wav_file.writeframes(packed_value)

    return str(output_path)


def create_test_wav_files(output_dir, num_files=1):
    """
    Create multiple test WAV files with different bit depths.

    Parameters
    ----------
    output_dir : str or Path
        Directory to save the WAV files
    num_files : int, default: 1
        Number of files to create

    Returns
    -------
    list
        Paths to the created WAV files
    """
    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    file_paths = []

    # Create 24-bit WAV files
    for i in range(num_files):
        file_path = output_dir / f"test_24bit_{i}.wav"
        create_24bit_wav_file(file_path)
        file_paths.append(str(file_path))

    return file_paths


if __name__ == "__main__":
    # Create a test directory in the current working directory
    test_dir = Path("test_wav_files")

    # Create test WAV files
    file_paths = create_test_wav_files(test_dir, num_files=2)

    print(f"Created {len(file_paths)} 24-bit WAV files:")
    for path in file_paths:
        print(f"  - {path}")



"""Abstract class defining the structure of all Extractor-based Interfaces."""

from abc import ABC
from typing import Optional

from .basetemporalalignmentinterface import BaseTemporalAlignmentInterface
from .tools import get_package


class BaseExtractorInterface(BaseTemporalAlignmentInterface, ABC):
    """
    Abstract class defining the structure of all Extractor-based Interfaces.
    """

    # Manually override any of these attributes in a subclass if needed.
    # Note that values set at the level of class definition are called upon import.
    ExtractorModuleName: Optional[str] = None
    ExtractorName: Optional[str] = None  # Defaults to __name__.replace("Interface", "Extractor").
    Extractor = None  # Class loads dynamically on first call to .get_extractor()

    @classmethod
    def get_extractor(cls):
        """
        Get the extractor class for this interface.

        Returns
        -------
        type
            The extractor class that will be used to read the data.
        """
        if cls.Extractor is not None:
            return cls.Extractor
        extractor_module = get_package(package_name=cls.ExtractorModuleName)
        extractor = getattr(
            extractor_module,
            cls.ExtractorName or cls.__name__.replace("Interface", "Extractor"),
        )
        cls.Extractor = extractor
        return extractor

    def __init__(self, **source_data):
        super().__init__(**source_data)
        self.extractor = self.get_extractor()
        self.extractor_kwargs = self._source_data_to_extractor_kwargs(source_data)
        self._extractor_instance = self.extractor(**self.extractor_kwargs)

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:
        """This functions maps the source_data to kwargs required to initialize the Extractor."""
        return source_data



from .roiextractors import (
    _check_if_imaging_fits_into_memory,
    get_nwb_imaging_metadata,
    get_nwb_segmentation_metadata,
    add_devices_to_nwbfile,
    add_fluorescence_traces_to_nwbfile,
    add_image_segmentation_to_nwbfile,
    add_imaging_to_nwbfile,
    add_imaging_plane_to_nwbfile,
    add_photon_series_to_nwbfile,
    add_plane_segmentation_to_nwbfile,
    add_segmentation_to_nwbfile,
    add_summary_images_to_nwbfile,
    write_imaging_to_nwbfile,
    write_segmentation_to_nwbfile,
)



from datetime import datetime
from typing import Optional
from warnings import warn

from pydantic import DirectoryPath, validate_call

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface


class OpenEphysLegacyRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary data interface for converting legacy Open Ephys data (.continuous files).

    Uses :py:class:`~spikeinterface.extractors.OpenEphysLegacyRecordingExtractor`.
    """

    display_name = "OpenEphys Legacy Recording"
    associated_suffixes = (".continuous", ".openephys", ".xml")
    info = "Interface for converting legacy OpenEphys recording data."

    @classmethod
    def get_stream_names(cls, folder_path: DirectoryPath) -> list[str]:
        """
        Get the names of available recording streams in the OpenEphys legacy folder.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to directory containing OpenEphys legacy files.

        Returns
        -------
        list of str
            The names of the available recording streams.
        """
        from spikeinterface.extractors import OpenEphysLegacyRecordingExtractor

        stream_names, _ = OpenEphysLegacyRecordingExtractor.get_streams(folder_path=folder_path)
        return stream_names

    @classmethod
    def get_source_schema(cls):
        """
        Compile input schema for the RecordingExtractor.

        Returns
        -------
        dict
            The JSON schema for the OpenEphys legacy recording interface source data,
            containing folder path and other configuration parameters. The schema
            inherits from the base recording extractor interface schema.
        """
        source_schema = super().get_source_schema()
        source_schema["properties"]["folder_path"][
            "description"
        ] = "Path to directory containing OpenEphys legacy files."

        return source_schema

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        stream_name: Optional[str] = None,
        block_index: Optional[int] = None,
        verbose: bool = False,
        es_key: str = "ElectricalSeries",
    ):
        """
        Initialize reading of OpenEphys legacy recording (.continuous files).

        See :py:class:`~spikeinterface.extractors.OpenEphysLegacyRecordingExtractor` for options.

        Parameters
        ----------
        folder_path : FolderPathType
            Path to OpenEphys directory.
        stream_name : str, optional
            The name of the recording stream.
        block_index : int, optional, default: None
            The index of the block to extract from the data.
        verbose : bool, default: Falseee
        es_key : str, default: "ElectricalSeries"
        """
        available_streams = self.get_stream_names(folder_path=folder_path)
        if len(available_streams) > 1 and stream_name is None:
            raise ValueError(
                "More than one stream is detected! "
                "Please specify which stream you wish to load with the `stream_name` argument. "
                "To see what streams are available, call "
                "`OpenEphysRecordingInterface.get_stream_names(folder_path=...)`."
            )
        if stream_name is not None and stream_name not in available_streams:
            raise ValueError(
                f"The selected stream '{stream_name}' is not in the available streams '{available_streams}'!"
            )

        super().__init__(
            folder_path=folder_path, stream_name=stream_name, block_index=block_index, verbose=verbose, es_key=es_key
        )

    def get_metadata(self):
        metadata = super().get_metadata()

        neo_reader = self.recording_extractor.neo_reader
        block_annotations = neo_reader.raw_annotations.get("blocks", [])
        if block_annotations:
            segment_annotations = block_annotations[0].get("segments", [])
            if segment_annotations:
                expected_date_format = "%d-%b-%Y %H%M%S"
                date_created = segment_annotations[0]["date_created"]
                date_created = date_created.strip("'")
                extracted_date, extracted_timestamp = date_created.split(" ")
                if len(extracted_timestamp) != len("%H%M%S"):
                    warn(
                        "The timestamp for starting time from openephys metadata is ambiguous "
                        f"('{extracted_timestamp}')! Only the date will be auto-populated in metadata. "
                        "Please update the timestamp manually to record this value with the highest known "
                        "temporal resolution."
                    )
                    session_start_time = datetime.strptime(extracted_date, "%d-%b-%Y")
                else:
                    session_start_time = datetime.strptime(date_created, expected_date_format)
                metadata["NWBFile"].update(session_start_time=session_start_time)

        return metadata



"""Interface for converting single or multiple images to NWB format."""

from pathlib import Path
from typing import List, Literal, Optional, Union

import numpy as np
from hdmf.data_utils import AbstractDataChunkIterator, DataChunk
from pynwb import NWBFile
from pynwb.base import Images
from pynwb.image import GrayscaleImage, RGBAImage, RGBImage

from ...basedatainterface import BaseDataInterface
from ...utils import DeepDict


class SingleImageIterator(AbstractDataChunkIterator):
    """Simple iterator to return a single image. This avoids loading the entire image into memory at initializing
    and instead loads it at writing time one by one"""

    def __init__(self, file_path: Union[str, Path]):
        self._file_path = Path(file_path)
        from PIL import Image

        # Get image information without loading the full image
        with Image.open(self._file_path) as img:
            self.image_mode = img.mode
            self._image_shape = img.size[::-1]  # PIL uses (width, height) instead of (height, width)
            self._max_shape = (None, None)

            self.number_of_bands = len(img.getbands())
            if self.number_of_bands > 1:
                self._image_shape += (self.number_of_bands,)
                self._max_shape += (self.number_of_bands,)

            # For LA mode, adjust shape to RGBA
            if self.image_mode == "LA":
                self._image_shape = self._image_shape[:-1] + (4,)
                self._max_shape = self._max_shape[:-1] + (4,)

            # Calculate file size in bytes
            self._size_bytes = self._file_path.stat().st_size
            # Calculate approximate memory size when loaded as numpy array
            self._memory_size = np.prod(self._image_shape) * np.dtype(float).itemsize

        self._images_returned = 0  # Number of images returned in __next__

    def _la_to_rgba(self, la_image: np.ndarray) -> np.ndarray:
        """Convert a Luminance-Alpha (LA) image to RGBA format without losing information."""
        if len(la_image.shape) != 3 or la_image.shape[2] != 2:
            raise ValueError("Input must be an LA image with shape (height, width, 2)")

        height, width, _ = la_image.shape
        rgba_image = np.zeros((height, width, 4), dtype=la_image.dtype)

        # Extract L and A channels
        l_channel = la_image[..., 0]
        a_channel = la_image[..., 1]

        # Copy L channel to R, G, and B channels
        rgba_image[..., 0] = l_channel  # Red
        rgba_image[..., 1] = l_channel  # Green
        rgba_image[..., 2] = l_channel  # Blue
        rgba_image[..., 3] = a_channel  # Alpha

        return rgba_image

    def __iter__(self):
        """Return the iterator object"""
        return self

    def __next__(self):
        """Return the DataChunk with the single full image"""
        from PIL import Image

        if self._images_returned == 0:
            data = np.asarray(Image.open(self._file_path))

            # Transform LA to RGBA if needed
            if self.image_mode == "LA":
                data = self._la_to_rgba(data)

            selection = (slice(None),) * data.ndim
            self._images_returned += 1
            return DataChunk(data=data, selection=selection)
        else:
            raise StopIteration

    def recommended_chunk_shape(self):
        """Recommend the chunk shape for the data array."""
        return self._image_shape

    def recommended_data_shape(self):
        """Recommend the initial shape for the data array."""
        return self._image_shape

    @property
    def dtype(self):
        """Define the data type of the array"""
        return np.dtype(float)

    @property
    def maxshape(self):
        """Property describing the maximum shape of the data array that is being iterated over"""
        return self._max_shape

    def __len__(self):
        return self._image_shape[0]

    @property
    def image_info(self):
        """Return dictionary with image information"""
        return {
            "file_size_bytes": self._size_bytes,
            "memory_size_bytes": self._memory_size,
            "shape": self._image_shape,
            "mode": self.image_mode,
            "bands": self.number_of_bands,
        }


class ImageInterface(BaseDataInterface):
    """Interface for converting single or multiple images to NWB format."""

    display_name = "Image Interface"
    keywords = ("image",)
    associated_suffixes = (".png", ".jpg", ".jpeg", ".tiff", ".tif", "webp")
    info = "Interface for converting single or multiple images to NWB format."

    # Mapping from PIL mode to NWB image class
    IMAGE_MODE_TO_NWB_TYPE_MAP = {
        "L": GrayscaleImage,
        "RGB": RGBImage,
        "RGBA": RGBAImage,
        "LA": RGBAImage,  # LA will be converted to RGBA
    }

    @classmethod
    def get_source_schema(cls) -> dict:
        """Return the schema for the source_data."""
        return dict(
            required=["file_paths"],
            properties=dict(
                file_paths=dict(
                    type="array",
                    items=dict(type="string"),
                    description="List of paths to image files to be converted",
                ),
                folder_path=dict(
                    type="string",
                    description="Path to folder containing images to be converted. Used if file_paths not provided.",
                ),
            ),
        )

    def __init__(
        self,
        file_paths: Optional[List[Union[str, Path]]] = None,
        folder_path: Optional[Union[str, Path]] = None,
        images_location: Literal["acquisition", "stimulus"] = "acquisition",
        verbose: bool = True,
    ):
        """
        Initialize the ImageInterface.

        Parameters
        ----------
        file_paths : list of Union[str, Path], optional
            List of paths to image files to be converted
        folder_path : Union[str, Path], optional
            Path to folder containing images to be converted. Used if file_paths not provided.
        images_location : Literal["acquisition", "stimulus"], default: "acquisition"
            Location to store images in the NWB file
        verbose : bool, default: True
            Whether to print status messages
        """
        if file_paths is None and folder_path is None:
            raise ValueError("Either file_paths or folder_path must be provided")

        if file_paths is not None and folder_path is not None:
            raise ValueError("Only one of file_paths or folder_path should be provided")

        self.file_paths = file_paths
        self.folder_path = folder_path
        self.images_location = images_location

        super().__init__(
            verbose=verbose,
            file_paths=file_paths,
            folder_path=folder_path,
            images_location=images_location,
        )

        # Process paths
        if folder_path is not None:
            folder = Path(folder_path)
            if not folder.exists():
                raise ValueError(f"Folder path {folder} does not exist")

            # Get all image files in folder
            file_paths = []
            for suffix in self.associated_suffixes:
                file_paths.extend(folder.glob(f"*{suffix}"))

            if not file_paths:
                raise ValueError(f"No image files found in {folder}")

        self.file_paths = [Path(p).resolve() for p in file_paths]

    def get_metadata(self) -> DeepDict:
        """Get metadata for the images."""
        metadata = super().get_metadata()

        # Add basic metadata about the images
        metadata["Images"] = dict(description="Images loaded through ImageInterface")

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[DeepDict] = None,
        container_name: str = "images",
    ) -> None:
        """
        Add the image data to an NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to add the images to
        metadata : dict, optional
            Metadata for the images
        container_name : str, default: "images"
            Name of the Images container
        """
        if metadata is None:
            metadata = self.get_metadata()

        # Create Images container
        images_container = Images(
            name=container_name,
            description=metadata.get("Images", {}).get("description", "Images loaded through ImageInterface"),
        )

        # Process each image
        for file_path in self.file_paths:
            # Create iterator for memory-efficient loading
            iterator = SingleImageIterator(file_path)

            # Get image name from file name
            image_name = Path(file_path).stem

            # Validate mode and get image class
            if iterator.image_mode not in self.IMAGE_MODE_TO_NWB_TYPE_MAP:
                raise ValueError(f"Unsupported image mode: {iterator.image_mode} for image {file_path.name}")

            nwb_image_class = self.IMAGE_MODE_TO_NWB_TYPE_MAP[iterator.image_mode]
            image_container = nwb_image_class(name=image_name, data=iterator)

            # Add to images container
            images_container.add_image(image_container)

        # Add images container to file
        if self.images_location == "acquisition":
            nwbfile.add_acquisition(images_container)
        else:
            nwbfile.add_stimulus(images_container)



from typing import Literal, Optional

from pydantic import DirectoryPath, FilePath
from pynwb import NWBFile

from ... import (
    BrukerTiffMultiPlaneImagingInterface,
    BrukerTiffSinglePlaneImagingInterface,
)
from ....nwbconverter import NWBConverter
from ....tools.nwb_helpers import make_or_load_nwbfile
from ....utils import get_json_schema_from_method_signature


class BrukerTiffMultiPlaneConverter(NWBConverter):
    """
    Converter class for Bruker imaging data with multiple channels and multiple planes.
    """

    display_name = "Bruker TIFF Imaging (multiple channels, multiple planes)"
    keywords = BrukerTiffMultiPlaneImagingInterface.keywords
    associated_suffixes = BrukerTiffMultiPlaneImagingInterface.associated_suffixes
    info = "Interface for handling all channels and all planes of Bruker imaging data."

    @classmethod
    def get_source_schema(cls):
        source_schema = get_json_schema_from_method_signature(cls)
        source_schema["properties"]["folder_path"][
            "description"
        ] = "The folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env)."
        return source_schema

    def get_conversion_options_schema(self) -> dict:
        """
        Get the schema for the conversion options.

        Returns
        -------
        dict
            The schema dictionary containing conversion options for the Bruker TIFF interface.
        """
        interface_name = list(self.data_interface_objects.keys())[0]
        return self.data_interface_objects[interface_name].get_conversion_options_schema()

    def __init__(
        self,
        folder_path: DirectoryPath,
        plane_separation_type: Literal["disjoint", "contiguous"] = None,
        verbose: bool = False,
    ):
        """
        Initializes the data interfaces for Bruker volumetric imaging data stream.

        Parameters
        ----------
        folder_path : DirectoryPath
            The path to the folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env).
        plane_separation_type: {'contiguous', 'disjoint'}
            Defines how to write volumetric imaging data. Use 'contiguous' to create the volumetric two photon series,
            and 'disjoint' to create separate imaging plane and two photon series for each plane.
        verbose : bool, default: False
            Controls verbosity.
        """
        self.verbose = verbose
        self.data_interface_objects = dict()

        if plane_separation_type is None or plane_separation_type not in ["disjoint", "contiguous"]:
            raise ValueError(
                "For volumetric imaging data the plane separation method must be one of 'disjoint' or 'contiguous'."
            )

        streams = BrukerTiffMultiPlaneImagingInterface.get_streams(
            folder_path=folder_path,
            plane_separation_type=plane_separation_type,
        )

        channel_streams = streams["channel_streams"]
        interface_name = "BrukerImaging"
        for channel_stream_name in channel_streams:
            plane_streams = streams["plane_streams"][channel_stream_name]
            for plane_stream in plane_streams:
                if len(plane_streams) > 1:
                    interface_name += plane_stream.replace("_", "")
                if plane_separation_type == "contiguous":
                    self.data_interface_objects[interface_name] = BrukerTiffMultiPlaneImagingInterface(
                        folder_path=folder_path,
                        stream_name=plane_stream,
                    )
                elif plane_separation_type == "disjoint":
                    self.data_interface_objects[interface_name] = BrukerTiffSinglePlaneImagingInterface(
                        folder_path=folder_path,
                        stream_name=plane_stream,
                    )

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata,
        stub_test: bool = False,
        stub_frames: int = 100,
    ):
        """
        Add data from multiple data interfaces to the specified NWBFile.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile object to which the data will be added.
        metadata : dict
            Metadata dictionary containing information to describe the data being added to the NWB file.
        stub_test : bool, optional
            If True, only a subset of the data (up to `stub_frames`) will be added for testing purposes. Default is False.
        stub_frames : int, optional
            The number of frames to include in the subset if `stub_test` is True. Default is 100.
        """
        for photon_series_index, (interface_name, data_interface) in enumerate(self.data_interface_objects.items()):
            data_interface.add_to_nwbfile(
                nwbfile=nwbfile,
                metadata=metadata,
                photon_series_index=photon_series_index,
                stub_test=stub_test,
                stub_frames=stub_frames,
            )

    def run_conversion(
        self,
        nwbfile_path: Optional[FilePath] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        stub_test: bool = False,
        stub_frames: int = 100,
    ) -> None:
        """
        Run the conversion process for the instantiated data interfaces and add data to the NWB file.

        Parameters
        ----------
        nwbfile_path : FilePath, optional
            Path where the NWB file will be written. If None, the file will be handled in-memory.
        nwbfile : NWBFile, optional
            An in-memory NWBFile object. If None, a new NWBFile object will be created.
        metadata : dict, optional
            Metadata dictionary for describing the NWB file. If None, it will be auto-generated using the `get_metadata()` method.
        overwrite : bool, optional
            If True, overwrites the existing NWB file at `nwbfile_path`. If False, appends to the file (default is False).
        stub_test : bool, optional
            If True, only a subset of the data (up to `stub_frames`) will be added for testing purposes, by default False.
        stub_frames : int, optional
            The number of frames to include in the subset if `stub_test` is True, by default 100.
        """
        if metadata is None:
            metadata = self.get_metadata()

        self.validate_metadata(metadata=metadata)

        self.temporally_align_data_interfaces()

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
        ) as nwbfile_out:
            self.add_to_nwbfile(nwbfile=nwbfile_out, metadata=metadata, stub_test=stub_test, stub_frames=stub_frames)


class BrukerTiffSinglePlaneConverter(NWBConverter):
    """
    Primary data interface class for converting Bruker imaging data with multiple channels and a single plane.
    """

    display_name = "Bruker TIFF Imaging (multiple channels, single plane)"
    keywords = BrukerTiffMultiPlaneImagingInterface.keywords
    associated_suffixes = BrukerTiffMultiPlaneImagingInterface.associated_suffixes
    info = "Interface for handling multiple channels of a single plane of Bruker imaging data."

    @classmethod
    def get_source_schema(cls):
        return get_json_schema_from_method_signature(cls)

    def get_conversion_options_schema(self) -> dict:
        """
        Get the schema for the conversion options.

        Returns
        -------
        dict
            The schema dictionary containing conversion options for the Bruker TIFF interface.
        """
        interface_name = list(self.data_interface_objects.keys())[0]
        return self.data_interface_objects[interface_name].get_conversion_options_schema()

    def __init__(
        self,
        folder_path: DirectoryPath,
        verbose: bool = False,
    ):
        """
        Initializes the data interfaces for Bruker imaging data stream.

        Parameters
        ----------
        folder_path : DirectoryPath
            The path to the folder that contains the Bruker TIF image files (.ome.tif) and configuration files (.xml, .env).
        verbose : bool, default: False
            Controls verbosity.
        """
        from roiextractors.extractors.tiffimagingextractors.brukertiffimagingextractor import (
            _determine_imaging_is_volumetric,
        )

        if _determine_imaging_is_volumetric(folder_path=folder_path):
            raise ValueError("For volumetric imaging data use BrukerTiffMultiPlaneConverter.")

        self.verbose = verbose
        self.data_interface_objects = dict()

        streams = BrukerTiffSinglePlaneImagingInterface.get_streams(folder_path=folder_path)
        channel_streams = streams["channel_streams"]
        interface_name = "BrukerImaging"
        for channel_stream_name in channel_streams:
            if len(channel_streams) > 1:
                interface_name += channel_stream_name.replace("_", "")
            self.data_interface_objects[interface_name] = BrukerTiffSinglePlaneImagingInterface(
                folder_path=folder_path,
                stream_name=channel_stream_name,
            )

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata,
        stub_test: bool = False,
        stub_frames: int = 100,
    ):
        """
        Add data from all instantiated data interfaces to the provided NWBFile.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile object to which the data will be added.
        metadata : dict
            Metadata dictionary containing information about the data to be added.
        stub_test : bool, optional
            If True, only a subset of the data (defined by `stub_frames`) will be added for testing purposes,
            by default False.
        stub_frames : int, optional
            The number of frames to include in the subset if `stub_test` is True, by default 100.
        """
        for photon_series_index, (interface_name, data_interface) in enumerate(self.data_interface_objects.items()):
            data_interface.add_to_nwbfile(
                nwbfile=nwbfile,
                metadata=metadata,
                photon_series_index=photon_series_index,
                stub_test=stub_test,
                stub_frames=stub_frames,
            )

    def run_conversion(
        self,
        nwbfile_path: Optional[FilePath] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        stub_test: bool = False,
        stub_frames: int = 100,
    ) -> None:
        """
        Run the NWB conversion process for all instantiated data interfaces.

        Parameters
        ----------
        nwbfile_path : FilePath, optional
            The file path where the NWB file will be written. If None, the file is handled in-memory.
        nwbfile : NWBFile, optional
            An existing in-memory NWBFile object. If None, a new NWBFile object will be created.
        metadata : dict, optional
            Metadata dictionary used to create or validate the NWBFile. If None, metadata is automatically generated.
        overwrite : bool, optional
            If True, the NWBFile at `nwbfile_path` is overwritten if it exists. If False (default), data is appended.
        stub_test : bool, optional
            If True, only a subset of the data (up to `stub_frames`) is used for testing purposes. By default False.
        stub_frames : int, optional
            The number of frames to include in the subset if `stub_test` is True. By default 100.
        """
        if metadata is None:
            metadata = self.get_metadata()

        self.validate_metadata(metadata=metadata)

        self.temporally_align_data_interfaces()

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
        ) as nwbfile_out:
            self.add_to_nwbfile(nwbfile=nwbfile_out, metadata=metadata, stub_test=stub_test, stub_frames=stub_frames)



"""DataInterfaces for SpikeGLX."""

import warnings
from pathlib import Path
from typing import Optional

import numpy as np
from pydantic import DirectoryPath, FilePath, validate_call

from .spikeglx_utils import (
    add_recording_extractor_properties,
    fetch_stream_id_for_spikelgx_file,
    get_device_metadata,
    get_session_start_time,
)
from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_json_schema_from_method_signature


class SpikeGLXRecordingInterface(BaseRecordingExtractorInterface):
    """
    Primary SpikeGLX interface for converting raw SpikeGLX data using a :py:class:`~spikeinterface.extractors.SpikeGLXRecordingExtractor`.
    """

    display_name = "SpikeGLX Recording"
    keywords = BaseRecordingExtractorInterface.keywords + ("Neuropixels",)
    associated_suffixes = (".imec{probe_index}", ".ap", ".lf", ".meta", ".bin")
    info = "Interface for SpikeGLX recording data."

    ExtractorName = "SpikeGLXRecordingExtractor"

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["x_pitch", "y_pitch"])
        source_schema["properties"]["file_path"]["description"] = "Path to SpikeGLX ap.bin or lf.bin file."
        return source_schema

    def _source_data_to_extractor_kwargs(self, source_data: dict) -> dict:

        extractor_kwargs = source_data.copy()
        extractor_kwargs["folder_path"] = self.folder_path
        extractor_kwargs["all_annotations"] = True
        extractor_kwargs["stream_id"] = self.stream_id
        return extractor_kwargs

    @validate_call
    def __init__(
        self,
        file_path: Optional[FilePath] = None,
        verbose: bool = False,
        es_key: Optional[str] = None,
        folder_path: Optional[DirectoryPath] = None,
        stream_id: Optional[str] = None,
    ):
        """
        Parameters
        ----------
        folder_path: DirectoryPath
            Folder path containing the binary files of the SpikeGLX recording.
        stream_id: str, optional
            Stream ID of the SpikeGLX recording.
            Examples are 'imec0.ap', 'imec0.lf', 'imec1.ap', 'imec1.lf', etc.
        file_path : FilePathType
            Path to .bin file. Point to .ap.bin for SpikeGLXRecordingInterface and .lf.bin for SpikeGLXLFPInterface.
        verbose : bool, default: False
            Whether to output verbose text.
        es_key : str, the key to access the metadata of the ElectricalSeries.
        """

        if stream_id == "nidq":
            raise ValueError(
                "SpikeGLXRecordingInterface is not designed to handle nidq files. Use SpikeGLXNIDQInterface instead"
            )

        if file_path is not None:
            warnings.warn(
                "file_path is deprecated and will be removed by the end of 2025. "
                "The first argument of this interface will be `folder_path` afterwards. "
                "Use folder_path and stream_id instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if file_path is not None and stream_id is None:
            self.stream_id = fetch_stream_id_for_spikelgx_file(file_path)
            self.folder_path = Path(file_path).parent
        else:
            self.stream_id = stream_id
            self.folder_path = Path(folder_path)

        super().__init__(
            folder_path=folder_path,
            verbose=verbose,
            es_key=es_key,
        )

        signal_info_key = (0, self.stream_id)  # Key format is (segment_index, stream_id)
        self._signals_info_dict = self.recording_extractor.neo_reader.signals_info_dict[signal_info_key]
        self.meta = self._signals_info_dict["meta"]

        if es_key is None:
            stream_kind = self._signals_info_dict["stream_kind"]  # ap or lf
            stream_kind_caps = stream_kind.upper()
            device = self._signals_info_dict["device"].capitalize()  # imec0, imec1, etc.

            electrical_series_name = f"ElectricalSeries{stream_kind_caps}"

            # Add imec{probe_index} to the electrical series name when there are multiple probes
            # or undefined, `typeImEnabled` is present in the meta of all the production probes
            self.probes_enabled_in_run = int(self.meta.get("typeImEnabled", 0))
            if self.probes_enabled_in_run != 1:
                electrical_series_name += f"{device}"

            self.es_key = electrical_series_name

        # Set electrodes properties
        add_recording_extractor_properties(self.recording_extractor)

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()
        session_start_time = get_session_start_time(self.meta)
        if session_start_time:
            metadata["NWBFile"]["session_start_time"] = session_start_time

        # Device metadata
        device = get_device_metadata(self.meta)

        # Should follow pattern 'Imec0', 'Imec1', etc.
        probe_name = self._signals_info_dict["device"].capitalize()
        device["name"] = f"Neuropixels{probe_name}"

        # Add groups metadata
        metadata["Ecephys"]["Device"] = [device]
        electrode_groups = [
            dict(
                name=group_name,
                description=f"A group representing probe/shank '{group_name}'.",
                location="unknown",
                device=device["name"],
            )
            for group_name in set(self.recording_extractor.get_property("group_name"))
        ]
        metadata["Ecephys"]["ElectrodeGroup"] = electrode_groups

        # Electrodes columns descriptions
        metadata["Ecephys"]["Electrodes"] = [
            dict(name="group_name", description="Name of the ElectrodeGroup this electrode is a part of."),
            dict(name="contact_shapes", description="The shape of the electrode"),
            dict(name="contact_ids", description="The id of the contact on the electrode"),
            dict(
                name="inter_sample_shift",
                description=(
                    "Array of relative phase shifts for each channel, with values ranging from 0 to 1, "
                    "representing the fractional delay within the sampling period due to sequential ADC."
                ),
            ),
        ]

        if self.recording_extractor.get_probe().get_shank_count() > 1:
            metadata["Ecephys"]["Electrodes"].append(
                dict(name="shank_ids", description="The shank id of the electrode")
            )

        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        new_recording = self.get_extractor()(
            folder_path=self.folder_path,
            stream_id=self.stream_id,
        )  # TODO: add generic method for aliasing from NeuroConv signature to SI init
        if self._number_of_segments == 1:
            return new_recording.get_times()
        else:
            return [
                new_recording.get_times(segment_index=segment_index)
                for segment_index in range(self._number_of_segments)
            ]



from pathlib import Path
from typing import Optional

from pydantic import DirectoryPath, validate_call

from .spikeglxdatainterface import SpikeGLXRecordingInterface
from .spikeglxnidqinterface import SpikeGLXNIDQInterface
from ....nwbconverter import ConverterPipe
from ....utils import get_json_schema_from_method_signature


class SpikeGLXConverterPipe(ConverterPipe):
    """
    The simplest, easiest to use class for converting all SpikeGLX data in a folder.

    Primary conversion class for handling multiple SpikeGLX data streams.
    """

    display_name = "SpikeGLX Converter"
    keywords = SpikeGLXRecordingInterface.keywords + SpikeGLXNIDQInterface.keywords
    associated_suffixes = SpikeGLXRecordingInterface.associated_suffixes + SpikeGLXNIDQInterface.associated_suffixes
    info = "Converter for multi-stream SpikeGLX recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        """
        Get the schema for the source arguments.

        Returns
        -------
        dict
            The schema dictionary containing input parameters and descriptions
            for initializing the SpikeGLX converter.
        """
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["streams"])
        source_schema["properties"]["folder_path"]["description"] = "Path to the folder containing SpikeGLX streams."
        return source_schema

    @classmethod
    def get_streams(cls, folder_path: DirectoryPath) -> list[str]:
        """
        Get the stream IDs available in the folder.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing SpikeGLX streams.

        Returns
        -------
        list of str
            The IDs of all available streams in the folder.
        """
        from spikeinterface.extractors import SpikeGLXRecordingExtractor

        # The first entry is the stream ids the second is the stream names
        return SpikeGLXRecordingExtractor.get_streams(folder_path=folder_path)[0]

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        streams: Optional[list[str]] = None,
        verbose: bool = False,
    ):
        """
        Read all data from multiple streams stored in the SpikeGLX format.

        This can include...
            (a) single-probe but multi-band such as AP+LF streams
            (b) multi-probe with multi-band
            (c) with or without the associated NIDQ channels

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to folder containing the NIDQ stream and subfolders containing each IMEC stream.
        streams : list of strings, optional
            A specific list of streams you wish to load.
            To see which streams are available, run `SpikeGLXConverter.get_streams(folder_path="path/to/spikeglx")`.
            By default, all available streams are loaded.
        verbose : bool, default: False
            Whether to output verbose text.
        """
        folder_path = Path(folder_path)

        streams_ids = streams or self.get_streams(folder_path=folder_path)

        data_interfaces = dict()

        nidq_streams = [stream_id for stream_id in streams_ids if stream_id == "nidq"]
        electrical_streams = [stream_id for stream_id in streams_ids if stream_id not in nidq_streams]
        for stream_id in electrical_streams:
            data_interfaces[stream_id] = SpikeGLXRecordingInterface(folder_path=folder_path, stream_id=stream_id)

        for stream_id in nidq_streams:
            data_interfaces[stream_id] = SpikeGLXNIDQInterface(folder_path=folder_path)

        super().__init__(data_interfaces=data_interfaces, verbose=verbose)

    def get_conversion_options_schema(self) -> dict:
        conversion_options_schema = super().get_conversion_options_schema()
        conversion_options_schema["properties"].update(
            {name: interface.get_conversion_options_schema() for name, interface in self.data_interface_objects.items()}
        )
        return conversion_options_schema



"""Collection of helper functions for deploying NeuroConv in EC2 Batch jobs on AWS."""

import os
import time
import uuid
import warnings
from typing import Optional

import boto3
from pydantic import FilePath, validate_call

from ._rclone_transfer_batch_job import rclone_transfer_batch_job
from ._submit_aws_batch_job import submit_aws_batch_job

_RETRY_STATES = ["RUNNABLE", "PENDING", "STARTING", "RUNNING"]


@validate_call
def deploy_neuroconv_batch_job(
    *,
    rclone_command: str,
    yaml_specification_file_path: FilePath,
    job_name: str,
    efs_volume_name: str,
    rclone_config_file_path: Optional[FilePath] = None,
    status_tracker_table_name: str = "neuroconv_batch_status_tracker",
    compute_environment_name: str = "neuroconv_batch_environment",
    job_queue_name: str = "neuroconv_batch_queue",
    job_definition_name: Optional[str] = None,
    minimum_worker_ram_in_gib: int = 16,  # Higher than previous recommendations for safer buffering room
    minimum_worker_cpus: int = 4,
    region: Optional[str] = None,
) -> dict[str, str]:
    """
    Submit a job to AWS Batch for processing.

    Requires AWS credentials saved to files in the `~/.aws/` folder or set as environment variables.

    Parameters
    ----------
    rclone_command : str
        The command to pass directly to Rclone running on the EC2 instance.
            E.g.: "rclone copy my_drive:testing_rclone /mnt/efs/source"
        Must move data from or to '/mnt/efs/source'.
    yaml_specification_file_path : FilePath
        The path to the YAML file containing the NeuroConv specification.
    job_name : str
        The name of the job to submit.
    efs_volume_name : str
        The name of an EFS volume to be created and attached to the job.
        The path exposed to the container will always be `/mnt/efs`.
    rclone_config_file_path : FilePath, optional
        The path to the Rclone configuration file to use for the job.
        If unspecified, method will attempt to find the file in `~/.rclone` and will raise an error if it cannot.
    status_tracker_table_name : str, default: "neuroconv_batch_status_tracker"
        The name of the DynamoDB table to use for tracking job status.
    compute_environment_name : str, default: "neuroconv_batch_environment"
        The name of the compute environment to use for the job.
    job_queue_name : str, default: "neuroconv_batch_queue"
        The name of the job queue to use for the job.
    job_definition_name : str, optional
        The name of the job definition to use for the job.
        If unspecified, a name starting with 'neuroconv_batch_' will be generated.
    minimum_worker_ram_in_gib : int, default: 4
        The minimum amount of base worker memory required to run this job.
        Determines the EC2 instance type selected by the automatic 'best fit' selector.
        Recommended to be several GiB to allow comfortable buffer space for data chunk iterators.
    minimum_worker_cpus : int, default: 4
        The minimum number of CPUs required to run this job.
        A minimum of 4 is required, even if only one will be used in the actual process.
    region : str, optional
        The AWS region to use for the job.
        If not provided, we will attempt to load the region from your local AWS configuration.
        If that file is not found on your system, we will default to "us-east-2", the location of the DANDI Archive.

    Returns
    -------
    info : dict
        A dictionary containing information about this AWS Batch job.

        info["rclone_job_submission_info"] is the return value of `neuroconv.tools.aws.rclone_transfer_batch_job`.
        info["neuroconv_job_submission_info"] is the return value of `neuroconv.tools.aws.submit_job`.
    """
    efs_volume_name = efs_volume_name or f"neuroconv_batch_efs_volume_{uuid.uuid4().hex[:4]}"
    region = region or "us-east-2"

    if "/mnt/efs/source" not in rclone_command:
        message = (
            f"The Rclone command '{rclone_command}' does not contain a reference to '/mnt/efs/source'. "
            "Without utilizing the EFS mount, the instance is unlikely to have enough local disk space. "
            "The subfolder 'source' is also required to eliminate ambiguity in the transfer process."
        )
        raise ValueError(message)

    rclone_job_name = f"{job_name}_rclone_transfer"
    rclone_job_submission_info = rclone_transfer_batch_job(
        rclone_command=rclone_command,
        job_name=rclone_job_name,
        efs_volume_name=efs_volume_name,
        rclone_config_file_path=rclone_config_file_path,
        region=region,
    )
    rclone_job_id = rclone_job_submission_info["job_submission_info"]["jobId"]

    # Give the EFS and other aspects time to spin up before submitting next dependent job
    # (Otherwise, good chance that duplicate EFS will be created)
    aws_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", None)
    aws_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", None)

    batch_client = boto3.client(
        service_name="batch",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )
    efs_client = boto3.client(
        service_name="efs",
        region_name=region,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
    )

    available_efs_volumes = efs_client.describe_file_systems()
    matching_efs_volumes = [
        file_system
        for file_system in available_efs_volumes["FileSystems"]
        for tag in file_system["Tags"]
        if tag["Key"] == "Name" and tag["Value"] == efs_volume_name
    ]
    max_iterations = 10
    iteration = 0
    while len(matching_efs_volumes) == 0 and iteration < max_iterations:
        iteration += 1
        time.sleep(30)

        matching_efs_volumes = [
            file_system
            for file_system in available_efs_volumes["FileSystems"]
            for tag in file_system["Tags"]
            if tag["Key"] == "Name" and tag["Value"] == efs_volume_name
        ]

    if len(matching_efs_volumes) == 0:
        message = f"Unable to create EFS volume '{efs_volume_name}' after {max_iterations} attempts!"
        raise ValueError(message)

    docker_image = "ghcr.io/catalystneuro/neuroconv_yaml_variable:latest"

    with open(file=yaml_specification_file_path, mode="r") as io:
        yaml_specification_file_stream = io.read()

    neuroconv_job_name = f"{job_name}_neuroconv_deployment"
    job_dependencies = [{"jobId": rclone_job_id, "type": "SEQUENTIAL"}]
    neuroconv_job_submission_info = submit_aws_batch_job(
        job_name=neuroconv_job_name,
        docker_image=docker_image,
        environment_variables={
            "NEUROCONV_YAML": yaml_specification_file_stream,
            "NEUROCONV_DATA_PATH": "/mnt/efs/source",
            # TODO: would prefer this to use subfolders for source and output, but need logic for YAML
            # related code to create them if missing (hard to send EFS this command directly)
            # (the code was included in this PR, but a release cycle needs to complete for the docker images before
            # it can be used here)
            # "NEUROCONV_OUTPUT_PATH": "/mnt/efs/output",
            "NEUROCONV_OUTPUT_PATH": "/mnt/efs",
        },
        efs_volume_name=efs_volume_name,
        job_dependencies=job_dependencies,
        status_tracker_table_name=status_tracker_table_name,
        compute_environment_name=compute_environment_name,
        job_queue_name=job_queue_name,
        job_definition_name=job_definition_name,
        minimum_worker_ram_in_gib=minimum_worker_ram_in_gib,
        minimum_worker_cpus=minimum_worker_cpus,
        region=region,
    )

    info = {
        "rclone_job_submission_info": rclone_job_submission_info,
        "neuroconv_job_submission_info": neuroconv_job_submission_info,
    }

    # TODO: would be better to spin up third dependent job to clean up EFS volume after neuroconv job completes
    neuroconv_job_id = neuroconv_job_submission_info["job_submission_info"]["jobId"]
    job = None
    max_retries = 60 * 12  # roughly 12 hours max runtime (aside from internet loss) for checking cleanup
    sleep_time = 60  # 1 minute
    retry = 0.0
    time.sleep(sleep_time)
    while retry < max_retries:
        job_description_response = batch_client.describe_jobs(jobs=[neuroconv_job_id])
        if job_description_response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            # sleep but only increment retry by a small amount
            # (really should only apply if internet connection is temporarily lost)
            retry += 0.1
            time.sleep(sleep_time)

        job = job_description_response["jobs"][0]
        if job["status"] in _RETRY_STATES:
            retry += 1.0
            time.sleep(sleep_time)
        elif job["status"] == "SUCCEEDED":
            break

    if retry >= max_retries:
        message = (
            "Maximum retries reached for checking job completion for automatic EFS cleanup! "
            "Please delete the EFS volume manually."
        )
        warnings.warn(message=message, stacklevel=2)

        return info

    # Cleanup EFS after job is complete - must clear mount targets first, then wait before deleting the volume
    efs_volumes = efs_client.describe_file_systems()
    matching_efs_volumes = [
        file_system
        for file_system in efs_volumes["FileSystems"]
        for tag in file_system["Tags"]
        if tag["Key"] == "Name" and tag["Value"] == efs_volume_name
    ]
    if len(matching_efs_volumes) != 1:
        message = (
            f"Expected to find exactly one EFS volume with name '{efs_volume_name}', "
            f"but found {len(matching_efs_volumes)}\n\n{matching_efs_volumes=}\n\n!"
            "You will have to delete these manually."
        )
        warnings.warn(message=message, stacklevel=2)

        return info

    efs_volume = matching_efs_volumes[0]
    efs_id = efs_volume["FileSystemId"]
    mount_targets = efs_client.describe_mount_targets(FileSystemId=efs_id)
    for mount_target in mount_targets["MountTargets"]:
        efs_client.delete_mount_target(MountTargetId=mount_target["MountTargetId"])

    time.sleep(sleep_time)
    efs_client.delete_file_system(FileSystemId=efs_id)

    return info



"""Collection of helper functions for assessing and performing automated data transfers related to AWS."""

import warnings
from typing import Optional

from pydantic import FilePath, validate_call

from ._submit_aws_batch_job import submit_aws_batch_job


@validate_call
def rclone_transfer_batch_job(
    *,
    rclone_command: str,
    job_name: str,
    efs_volume_name: str,
    rclone_config_file_path: Optional[FilePath] = None,
    status_tracker_table_name: str = "neuroconv_batch_status_tracker",
    compute_environment_name: str = "neuroconv_batch_environment",
    job_queue_name: str = "neuroconv_batch_queue",
    job_definition_name: Optional[str] = None,
    minimum_worker_ram_in_gib: int = 4,
    minimum_worker_cpus: int = 4,
    submission_id: Optional[str] = None,
    region: Optional[str] = None,
) -> dict[str, str]:
    """
    Submit a job to AWS Batch for processing.

    Requires AWS credentials saved to files in the `~/.aws/` folder or set as environment variables.

    Parameters
    ----------
    rclone_command : str
        The command to pass directly to Rclone running on the EC2 instance.
            E.g.: "rclone copy my_drive:testing_rclone /mnt/efs"
        Must move data from or to '/mnt/efs'.
    job_name : str
        The name of the job to submit.
    efs_volume_name : str
        The name of an EFS volume to be created and attached to the job.
        The path exposed to the container will always be `/mnt/efs`.
    rclone_config_file_path : FilePath, optional
        The path to the Rclone configuration file to use for the job.
        If unspecified, method will attempt to find the file in `~/.rclone` and will raise an error if it cannot.
    status_tracker_table_name : str, default: "neuroconv_batch_status_tracker"
        The name of the DynamoDB table to use for tracking job status.
    compute_environment_name : str, default: "neuroconv_batch_environment"
        The name of the compute environment to use for the job.
    job_queue_name : str, default: "neuroconv_batch_queue"
        The name of the job queue to use for the job.
    job_definition_name : str, optional
        The name of the job definition to use for the job.
        If unspecified, a name starting with 'neuroconv_batch_' will be generated.
    minimum_worker_ram_in_gib : int, default: 4
        The minimum amount of base worker memory required to run this job.
        Determines the EC2 instance type selected by the automatic 'best fit' selector.
        Recommended to be several GiB to allow comfortable buffer space for data chunk iterators.
    minimum_worker_cpus : int, default: 4
        The minimum number of CPUs required to run this job.
        A minimum of 4 is required, even if only one will be used in the actual process.
    submission_id : str, optional
        The unique ID to pair with this job submission when tracking the status via DynamoDB.
        Defaults to a random UUID4.
    region : str, optional
        The AWS region to use for the job.
        If not provided, we will attempt to load the region from your local AWS configuration.
        If that file is not found on your system, we will default to "us-east-2", the location of the DANDI Archive.

    Returns
    -------
    info : dict
        A dictionary containing information about this AWS Batch job.

        info["job_submission_info"] is the return value of `boto3.client.submit_job` which contains the job ID.
        info["table_submission_info"] is the initial row data inserted into the DynamoDB status tracking table.
    """
    docker_image = "ghcr.io/catalystneuro/rclone_with_config:latest"

    if "/mnt/efs" not in rclone_command:
        message = (
            f"The Rclone command '{rclone_command}' does not contain a reference to '/mnt/efs'. "
            "Without utilizing the EFS mount, the instance is unlikely to have enough local disk space."
        )
        warnings.warn(message=message, stacklevel=2)

    rclone_config_file_path = rclone_config_file_path or pathlib.Path.home() / ".rclone" / "rclone.conf"
    if not rclone_config_file_path.exists():
        raise FileNotFoundError(
            f"Rclone configuration file not found at: {rclone_config_file_path}! "
            "Please check that `rclone config` successfully created the file."
        )
    with open(file=rclone_config_file_path, mode="r") as io:
        rclone_config_file_stream = io.read()

    region = region or "us-east-2"

    info = submit_aws_batch_job(
        job_name=job_name,
        docker_image=docker_image,
        environment_variables={"RCLONE_CONFIG": rclone_config_file_stream, "RCLONE_COMMAND": rclone_command},
        efs_volume_name=efs_volume_name,
        status_tracker_table_name=status_tracker_table_name,
        compute_environment_name=compute_environment_name,
        job_queue_name=job_queue_name,
        job_definition_name=job_definition_name,
        minimum_worker_ram_in_gib=minimum_worker_ram_in_gib,
        minimum_worker_cpus=minimum_worker_cpus,
        submission_id=submission_id,
        region=region,
    )

    return info



"""Author: Ben Dichter."""

from typing import Literal, Optional

import numpy as np
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ophys import ImagingPlane, OnePhotonSeries, TwoPhotonSeries

from ...baseextractorinterface import BaseExtractorInterface
from ...utils import (
    DeepDict,
    dict_deep_update,
    fill_defaults,
    get_base_schema,
    get_schema_from_hdmf_class,
)


class BaseImagingExtractorInterface(BaseExtractorInterface):
    """Parent class for all ImagingExtractorInterfaces."""

    keywords = (
        "ophys",
        "optical electrophysiology",
        "fluorescence",
        "microscopy",
        "two photon",
        "one photon",
        "voltage imaging",
        "calcium imaging",
    )

    ExtractorModuleName = "roiextractors"

    def __init__(
        self,
        verbose: bool = False,
        photon_series_type: Literal["OnePhotonSeries", "TwoPhotonSeries"] = "TwoPhotonSeries",
        **source_data,
    ):
        super().__init__(**source_data)
        self.imaging_extractor = self._extractor_instance
        self.verbose = verbose
        self.photon_series_type = photon_series_type

    def get_metadata_schema(
        self,
    ) -> dict:
        """
        Retrieve the metadata schema for the optical physiology (Ophys) data.

        Returns
        -------
        dict
            The metadata schema dictionary containing definitions for Device, ImagingPlane,
            and either OnePhotonSeries or TwoPhotonSeries based on the photon_series_type.
        """

        metadata_schema = super().get_metadata_schema()

        metadata_schema["required"] = ["Ophys"]

        # Initiate Ophys metadata
        metadata_schema["properties"]["Ophys"] = get_base_schema(tag="Ophys")
        metadata_schema["properties"]["Ophys"]["required"] = ["Device", "ImagingPlane", self.photon_series_type]
        metadata_schema["properties"]["Ophys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Ophys/definitions/Device"}),
            ImagingPlane=dict(type="array", minItems=1, items={"$ref": "#/properties/Ophys/definitions/ImagingPlane"}),
        )
        metadata_schema["properties"]["Ophys"]["properties"].update(
            {
                self.photon_series_type: dict(
                    type="array",
                    minItems=1,
                    items={"$ref": f"#/properties/Ophys/definitions/{self.photon_series_type}"},
                ),
            }
        )

        # Schema definition for arrays

        imaging_plane_schema = get_schema_from_hdmf_class(ImagingPlane)
        imaging_plane_schema["properties"]["optical_channel"].pop("maxItems")
        metadata_schema["properties"]["Ophys"]["definitions"] = dict(
            Device=get_schema_from_hdmf_class(Device),
            ImagingPlane=imaging_plane_schema,
        )
        photon_series = dict(
            OnePhotonSeries=OnePhotonSeries,
            TwoPhotonSeries=TwoPhotonSeries,
        )[self.photon_series_type]
        metadata_schema["properties"]["Ophys"]["definitions"].update(
            {
                self.photon_series_type: get_schema_from_hdmf_class(photon_series),
            }
        )

        fill_defaults(metadata_schema, self.get_metadata())
        return metadata_schema

    def get_metadata(
        self,
    ) -> DeepDict:
        """
        Retrieve the metadata for the imaging data.

        Returns
        -------
        DeepDict
            Dictionary containing metadata including device information, imaging plane details,
            and photon series configuration.
        """

        from ...tools.roiextractors import get_nwb_imaging_metadata

        metadata = super().get_metadata()
        default_metadata = get_nwb_imaging_metadata(self.imaging_extractor, photon_series_type=self.photon_series_type)
        metadata = dict_deep_update(default_metadata, metadata)

        # fix troublesome data types
        if "TwoPhotonSeries" in metadata["Ophys"]:
            for two_photon_series in metadata["Ophys"]["TwoPhotonSeries"]:
                if "dimension" in two_photon_series:
                    two_photon_series["dimension"] = list(two_photon_series["dimension"])
                if "rate" in two_photon_series:
                    two_photon_series["rate"] = float(two_photon_series["rate"])
        return metadata

    def get_original_timestamps(self) -> np.ndarray:
        reinitialized_extractor = self.get_extractor()(**self.extractor_kwargs)
        return reinitialized_extractor.frame_to_time(frames=np.arange(stop=reinitialized_extractor.get_num_frames()))

    def get_timestamps(self) -> np.ndarray:
        return self.imaging_extractor.frame_to_time(frames=np.arange(stop=self.imaging_extractor.get_num_frames()))

    def set_aligned_timestamps(self, aligned_timestamps: np.ndarray):
        self.imaging_extractor.set_times(times=aligned_timestamps)

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        photon_series_type: Literal["TwoPhotonSeries", "OnePhotonSeries"] = "TwoPhotonSeries",
        photon_series_index: int = 0,
        parent_container: Literal["acquisition", "processing/ophys"] = "acquisition",
        stub_test: bool = False,
        stub_frames: int = 100,
        always_write_timestamps: bool = False,
    ):
        """
        Add imaging data to the NWB file

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file where the imaging data will be added.
        metadata : dict, optional
            Metadata for the NWBFile, by default None.
        photon_series_type : {"TwoPhotonSeries", "OnePhotonSeries"}, optional
            The type of photon series to be added, by default "TwoPhotonSeries".
        photon_series_index : int, optional
            The index of the photon series in the provided imaging data, by default 0.
        parent_container : {"acquisition", "processing/ophys"}, optional
            Specifies the parent container to which the photon series should be added, either as part of "acquisition" or
            under the "processing/ophys" module, by default "acquisition".
        stub_test : bool, optional
            If True, only writes a small subset of frames for testing purposes, by default False.
        stub_frames : int, optional
            The number of frames to write when stub_test is True. Will use min(stub_frames, total_frames) to avoid
            exceeding available frames, by default 100.
        """

        from ...tools.roiextractors import add_imaging_to_nwbfile

        if stub_test:
            stub_frames = min([stub_frames, self.imaging_extractor.get_num_frames()])
            imaging_extractor = self.imaging_extractor.frame_slice(start_frame=0, end_frame=stub_frames)
        else:
            imaging_extractor = self.imaging_extractor

        metadata = metadata or self.get_metadata()

        add_imaging_to_nwbfile(
            imaging=imaging_extractor,
            nwbfile=nwbfile,
            metadata=metadata,
            photon_series_type=photon_series_type,
            photon_series_index=photon_series_index,
            parent_container=parent_container,
            always_write_timestamps=always_write_timestamps,
        )



import collections.abc
import inspect
import json
import warnings
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Optional

import docstring_parser
import hdmf.data_utils
import numpy as np
import pydantic
import pynwb
from jsonschema import validate
from pynwb.device import Device
from pynwb.icephys import IntracellularElectrode


class _GenericNeuroconvEncoder(json.JSONEncoder):
    """Generic JSON encoder for NeuroConv data."""

    def default(self, obj):
        """
        Serialize custom data types to JSON. This overwrites the default method of the JSONEncoder class.
        """
        # Over-write behaviors for datetime object
        if isinstance(obj, datetime):
            return obj.isoformat()

        # Transform numpy generic integers and floats to python ints floats
        if isinstance(obj, np.generic):
            return obj.item()

        # Numpy arrays should be converted to lists
        if isinstance(obj, np.ndarray):
            return obj.tolist()

        # Over-write behaviors for Paths
        if isinstance(obj, Path):
            return str(obj)

        # The base-class handles it
        return super().default(obj)


class _NWBMetaDataEncoder(_GenericNeuroconvEncoder):
    """
    Custom JSON encoder for NWB metadata.
    """


class _NWBSourceDataEncoder(_GenericNeuroconvEncoder):
    """
    Custom JSON encoder for data interface source data (i.e. kwargs).
    """


class _NWBConversionOptionsEncoder(_GenericNeuroconvEncoder):
    """
    Custom JSON encoder for conversion options of the data interfaces and converters (i.e. kwargs).
    """


# This is used in the Guide so we will keep it public.
NWBMetaDataEncoder = _NWBMetaDataEncoder


def get_base_schema(
    tag: Optional[str] = None,
    root: bool = False,
    id_: Optional[str] = None,
    required: Optional[list[str]] = None,
    properties: Optional[dict] = None,
    **kwargs,
) -> dict:
    """
    Return the base schema used for all other schemas.

    Parameters
    ----------
    tag : str, optional
        Tag to identify the schema.
    root : bool, default: False
        Whether this schema is a root schema.
    id_ : str, optional
        Schema identifier.
    required : list of str, optional
        List of required property names.
    properties : dict, optional
        Dictionary of property definitions.
    **kwargs
        Additional schema properties.

    Returns
    -------
    dict
        Base JSON schema with the following structure:
        {
            "required": List of required properties (empty if not provided)
            "properties": Dictionary of property definitions (empty if not provided)
            "type": "object"
            "additionalProperties": False
            "tag": Optional tag if provided
            "$schema": Schema version if root is True
            "$id": Schema ID if provided
            **kwargs: Any additional properties
        }
    """
    base_schema = dict(
        required=required or [],
        properties=properties or {},
        type="object",
        additionalProperties=False,
    )
    if tag is not None:
        base_schema.update(tag=tag)
    if root:
        base_schema.update({"$schema": "http://json-schema.org/draft-07/schema#"})
    if id_ is not None:
        base_schema.update({"$id": id_})
    base_schema.update(**kwargs)
    return base_schema


def get_json_schema_from_method_signature(method: Callable, exclude: Optional[list[str]] = None) -> dict:
    """
    Get the equivalent JSON schema for a signature of a method.

    Also uses `docstring_parser` (NumPy style) to attempt to find descriptions for the arguments.

    Parameters
    ----------
    method : callable
        The method to generate the JSON schema from.
    exclude : list of str, optional
        List of arguments to exclude from the schema generation.
        Always includes 'self' and 'cls'.

    Returns
    -------
    json_schema : dict
        The JSON schema corresponding to the method signature.
    """
    exclude = exclude or []
    exclude += ["self", "cls"]

    split_qualname = method.__qualname__.split(".")[-2:]
    method_display = ".".join(split_qualname) if "<" not in split_qualname[0] else method.__name__

    signature = inspect.signature(obj=method)
    parameters = signature.parameters
    additional_properties = False
    arguments_to_annotations = {}
    for argument_name in parameters:
        if argument_name in exclude:
            continue

        parameter = parameters[argument_name]

        if parameter.kind == inspect.Parameter.VAR_KEYWORD:  # Skip all **{...} usage
            additional_properties = True
            continue

        # Raise error if the type annotation is missing as a json schema cannot be generated in that case
        if parameter.annotation is inspect._empty:
            raise TypeError(
                f"Parameter '{argument_name}' in method '{method_display}' is missing a type annotation. "
                f"Either add a type annotation for '{argument_name}' or add it to the exclude list."
            )

        # Pydantic uses ellipsis for required
        pydantic_default = ... if parameter.default is inspect._empty else parameter.default
        arguments_to_annotations.update({argument_name: (parameter.annotation, pydantic_default)})

    # The ConfigDict is required to support custom types like NumPy arrays
    model = pydantic.create_model(
        "_TempModel", __config__=pydantic.ConfigDict(arbitrary_types_allowed=True), **arguments_to_annotations
    )

    temp_json_schema = model.model_json_schema()

    # We never used to include titles in the lower schema layers
    # But Pydantic does automatically
    json_schema = _copy_without_title_keys(temp_json_schema)

    # Pydantic does not make determinations on additionalProperties
    json_schema["additionalProperties"] = additional_properties

    # Attempt to find descriptions within the docstring of the method
    parsed_docstring = docstring_parser.parse(method.__doc__)
    for parameter_in_docstring in parsed_docstring.params:
        if parameter_in_docstring.arg_name in exclude:
            continue

        if parameter_in_docstring.arg_name not in json_schema["properties"]:
            message = (
                f"The argument_name '{parameter_in_docstring.arg_name}' from the docstring of method "
                f"'{method_display}' does not occur in the signature, possibly due to a typo."
            )
            warnings.warn(message=message, stacklevel=2)
            continue

        if parameter_in_docstring.description is not None:
            json_schema["properties"][parameter_in_docstring.arg_name].update(
                description=parameter_in_docstring.description
            )
    # TODO: could also add Field support for more direct control over docstrings (and enhanced validation conditions)

    return json_schema


def _copy_without_title_keys(d: Any, /) -> Optional[dict]:
    if not isinstance(d, dict):
        return d

    return {key: _copy_without_title_keys(value) for key, value in d.items() if key != "title"}


def fill_defaults(schema: dict, defaults: dict, overwrite: bool = True):
    """
    Insert the values of the defaults dict as default values in the schema in place.

    Parameters
    ----------
    schema: dict
    defaults: dict
    overwrite: bool
    """
    # patternProperties introduced with the CsvTimeIntervalsInterface
    # caused issue with NWBConverter.get_metadata_schema() call leading here
    properties_reference = "properties"
    if properties_reference not in schema and "patternProperties" in schema:
        properties_reference = "patternProperties"

    for key, val in schema[properties_reference].items():
        if key in defaults:
            if val["type"] == "object":
                fill_defaults(val, defaults[key], overwrite=overwrite)
            else:
                if overwrite or ("default" not in val):
                    val["default"] = defaults[key]


def unroot_schema(schema: dict):
    """
    Modify a json-schema dictionary to make it not root.

    Parameters
    ----------
    schema: dict
    """
    terms = ("required", "properties", "type", "additionalProperties", "title", "description")
    return {k: v for k, v in schema.items() if k in terms}


def _is_member(types, target_types):
    if not isinstance(target_types, tuple):
        target_types = (target_types,)
    if not isinstance(types, tuple):
        types = (types,)
    return any(t in target_types for t in types)


def get_schema_from_hdmf_class(hdmf_class):
    """
    Get metadata schema from hdmf class.

    Parameters
    ----------
    hdmf_class : type
        The HDMF class to generate a schema from.

    Returns
    -------
    dict
        JSON schema derived from the HDMF class, containing:
        - tag: Full class path (module.name)
        - required: List of required fields
        - properties: Dictionary of field definitions including types and descriptions
        - additionalProperties: Whether extra fields are allowed
    """
    schema = get_base_schema()
    schema["tag"] = hdmf_class.__module__ + "." + hdmf_class.__name__

    # Detect child-like (as opposed to link) fields
    pynwb_children_fields = [f["name"] for f in hdmf_class.get_fields_conf() if f.get("child", False)]
    # For MultiContainerInterface
    if hasattr(hdmf_class, "__clsconf__"):
        pynwb_children_fields.append(hdmf_class.__clsconf__["attr"])
    # Temporary solution before this is solved: https://github.com/hdmf-dev/hdmf/issues/475
    if "device" in pynwb_children_fields:
        pynwb_children_fields.remove("device")
    docval = hdmf_class.__init__.__docval__
    for docval_arg in docval["args"]:
        arg_name = docval_arg["name"]
        arg_type = docval_arg["type"]

        schema_val = dict(description=docval_arg["doc"])

        if arg_name == "name":
            schema_val.update(pattern="^[^/]*$")

        if _is_member(arg_type, (float, int, "float", "int")):
            schema_val.update(type="number")
        elif _is_member(arg_type, str):
            schema_val.update(type="string")
        elif _is_member(arg_type, collections.abc.Iterable):
            schema_val.update(type="array")
        elif isinstance(arg_type, tuple) and (np.ndarray in arg_type and hdmf.data_utils.DataIO not in arg_type):
            # extend type array without including type where DataIO in tuple
            schema_val.update(type="array")
        elif _is_member(arg_type, datetime):
            schema_val.update(type="string", format="date-time")
        elif _is_member(arg_type, (pynwb.base.TimeSeries, pynwb.ophys.PlaneSegmentation)):
            continue
        else:
            if not isinstance(arg_type, tuple):
                docval_arg_type = [arg_type]
            else:
                docval_arg_type = arg_type
            # if another nwb object (or list of nwb objects)
            if any([hasattr(t, "__nwbfields__") for t in docval_arg_type]):
                is_nwb = [hasattr(t, "__nwbfields__") for t in docval_arg_type]
                item = docval_arg_type[np.where(is_nwb)[0][0]]
                # if it is child
                if arg_name in pynwb_children_fields:
                    items = get_schema_from_hdmf_class(item)
                    schema_val.update(type="array", items=items, minItems=1, maxItems=1)
                # if it is a link
                else:
                    target = item.__module__ + "." + item.__name__
                    schema_val.update(type="string", target=target)
            else:
                continue
        # Check for default arguments
        if "default" in docval_arg:
            if docval_arg["default"] is not None:
                schema_val.update(default=docval_arg["default"])
        else:
            schema["required"].append(arg_name)
        schema["properties"][arg_name] = schema_val
    if "allow_extra" in docval:
        schema["additionalProperties"] = docval["allow_extra"]
    return schema


def get_metadata_schema_for_icephys() -> dict:
    """
    Returns the metadata schema for icephys data.

    Returns
    -------
    dict
        The metadata schema for icephys data, containing definitions for Device,
        Electrode, and Session configurations.
    """
    schema = get_base_schema(tag="Icephys")
    schema["required"] = ["Device", "Electrodes"]
    schema["properties"] = dict(
        Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Icephys/definitions/Device"}),
        Electrodes=dict(
            type="array",
            minItems=1,
            items={"$ref": "#/properties/Icephys/definitions/Electrode"},
        ),
        Sessions=dict(
            type="array",
            minItems=1,
            items={"$ref": "#/properties/Icephys/definitions/Sessions"},
        ),
    )

    schema["definitions"] = dict(
        Device=get_schema_from_hdmf_class(Device),
        Electrode=get_schema_from_hdmf_class(IntracellularElectrode),
        Sessions=dict(
            name={"type": "string", "description": "Session name."},
            relative_session_start_time={
                "type": "number",
                "description": "the start time of the sessions in seconds, relative to the absolute start time",
            },
            icephys_experiment_type={
                "type": "string",
                "description": "Icephys experiment type. Allowed types are: voltage_clamp, current_clamp and izero",
            },
            stimulus_type={
                "type": "string",
                "description": "Description of the type pf stimulus, e.g. Square current clamp.",
            },
            recordings=dict(
                type="array",
                minItems=1,
                items={"$ref": "#/properties/Icephys/definitions/SessionsRecordings"},
            ),
        ),
        SessionsRecordings=dict(
            intracellular_recordings_table_ind={"type": "number", "description": ""},
            simultaneous_recordings_table_ind={"type": "number", "description": ""},
            sequential_recordings_table_ind={"type": "number", "description": ""},
        ),
    )

    return schema


def validate_metadata(metadata: dict[str, dict], schema: dict[str, dict], verbose: bool = False):
    """Validate metadata against a schema."""
    encoder = _NWBMetaDataEncoder()
    # The encoder produces a serialized object, so we deserialized it for comparison

    serialized_metadata = encoder.encode(metadata)
    decoded_metadata = json.loads(serialized_metadata)
    validate(instance=decoded_metadata, schema=schema)
    if verbose:
        print("Metadata is valid!")



import warnings
from pathlib import Path
from typing import Literal, Optional

import numpy as np
from pydantic import ConfigDict, DirectoryPath, FilePath, validate_call
from pynwb import NWBFile
from pynwb.base import TimeSeries

from .spikeglx_utils import get_session_start_time
from ....basedatainterface import BaseDataInterface
from ....tools.signal_processing import get_rising_frames_from_ttl
from ....tools.spikeinterface.spikeinterface import _recording_traces_to_hdmf_iterator
from ....utils import (
    calculate_regular_series_rate,
    get_json_schema_from_method_signature,
)


class SpikeGLXNIDQInterface(BaseDataInterface):
    """Primary data interface class for converting the high-pass (ap) SpikeGLX format."""

    display_name = "NIDQ Recording"
    keywords = ("Neuropixels", "nidq", "NIDQ", "SpikeGLX")
    associated_suffixes = (".nidq", ".meta", ".bin")
    info = "Interface for NIDQ board recording data."

    @classmethod
    def get_source_schema(cls) -> dict:
        source_schema = get_json_schema_from_method_signature(method=cls.__init__, exclude=["x_pitch", "y_pitch"])
        source_schema["properties"]["file_path"]["description"] = "Path to SpikeGLX .nidq file."
        return source_schema

    @validate_call(config=ConfigDict(arbitrary_types_allowed=True))
    def __init__(
        self,
        file_path: Optional[FilePath] = None,
        verbose: bool = False,
        load_sync_channel: Optional[bool] = None,
        es_key: str = "ElectricalSeriesNIDQ",
        folder_path: Optional[DirectoryPath] = None,
    ):
        """
        Read channel data from the NIDQ board for the SpikeGLX recording.

        Useful for synchronizing multiple data streams into the common time basis of the SpikeGLX system.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing the .nidq.bin file.
        file_path : FilePathType
            Path to .nidq.bin file.
        verbose : bool, default: False
            Whether to output verbose text.
        es_key : str, default: "ElectricalSeriesNIDQ"
        """

        if load_sync_channel is not None:

            warnings.warn(
                "The 'load_sync_channel' parameter is deprecated and will be removed in June 2025. "
                "The sync channel data is only available the raw files of spikeglx`.",
                DeprecationWarning,
                stacklevel=2,
            )

        if file_path is not None:
            warnings.warn(
                "file_path is deprecated and will be removed by the end of 2025. "
                "The first argument of this interface will be `folder_path` afterwards. "
                "Use folder_path and stream_id instead.",
                DeprecationWarning,
                stacklevel=2,
            )

        if file_path is None and folder_path is None:
            raise ValueError("Either 'file_path' or 'folder_path' must be provided.")

        if file_path is not None:
            file_path = Path(file_path)
            self.folder_path = file_path.parent

        if folder_path is not None:
            self.folder_path = Path(folder_path)

        from spikeinterface.extractors import SpikeGLXRecordingExtractor

        self.recording_extractor = SpikeGLXRecordingExtractor(
            folder_path=self.folder_path,
            stream_id="nidq",
            all_annotations=True,
        )

        channel_ids = self.recording_extractor.get_channel_ids()
        analog_channel_signatures = ["XA", "MA"]
        self.analog_channel_ids = [ch for ch in channel_ids if "XA" in ch or "MA" in ch]
        self.has_analog_channels = len(self.analog_channel_ids) > 0
        self.has_digital_channels = len(self.analog_channel_ids) < len(channel_ids)
        if self.has_digital_channels:
            import ndx_events  # noqa: F401
            from spikeinterface.extractors import SpikeGLXEventExtractor

            self.event_extractor = SpikeGLXEventExtractor(folder_path=self.folder_path)

        super().__init__(
            verbose=verbose,
            load_sync_channel=load_sync_channel,
            es_key=es_key,
            folder_path=self.folder_path,
            file_path=file_path,
        )

        self.subset_channels = None

        signal_info_key = (0, "nidq")  # Key format is (segment_index, stream_id)
        self._signals_info_dict = self.recording_extractor.neo_reader.signals_info_dict[signal_info_key]
        self.meta = self._signals_info_dict["meta"]

    def get_metadata(self) -> dict:
        metadata = super().get_metadata()

        session_start_time = get_session_start_time(self.meta)
        if session_start_time:
            metadata["NWBFile"]["session_start_time"] = session_start_time

        # Device metadata
        device = dict(
            name="NIDQBoard",
            description="A NIDQ board used in conjunction with SpikeGLX.",
            manufacturer="National Instruments",
        )

        metadata["Devices"] = [device]

        return metadata

    def get_channel_names(self) -> list[str]:
        """
        Get a list of channel names from the recording extractor.

        Returns
        -------
        list of str
            The names of all channels in the NIDQ recording.
        """
        return list(self.recording_extractor.get_channel_ids())

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        stub_test: bool = False,
        starting_time: Optional[float] = None,
        write_as: Literal["raw", "lfp", "processed"] = "raw",
        write_electrical_series: bool = True,
        iterator_type: Optional[str] = "v2",
        iterator_opts: Optional[dict] = None,
        always_write_timestamps: bool = False,
    ):
        """
        Add NIDQ board data to an NWB file, including both analog and digital channels if present.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to which the NIDQ data will be added
        metadata : Optional[dict], default: None
            Metadata dictionary with device information. If None, uses default metadata
        stub_test : bool, default: False
            If True, only writes a small amount of data for testing
        starting_time : Optional[float], default: None
            DEPRECATED: Will be removed in June 2025. Starting time offset for the TimeSeries
        write_as : Literal["raw", "lfp", "processed"], default: "raw"
            DEPRECATED: Will be removed in June 2025. Specifies how to write the data
        write_electrical_series : bool, default: True
            DEPRECATED: Will be removed in June 2025. Whether to write electrical series data
        iterator_type : Optional[str], default: "v2"
            Type of iterator to use for data streaming
        iterator_opts : Optional[dict], default: None
            Additional options for the iterator
        always_write_timestamps : bool, default: False
            If True, always writes timestamps instead of using sampling rate
        """

        if starting_time is not None:
            warnings.warn(
                "The 'starting_time' parameter is deprecated and will be removed in June 2025. "
                "Use the time alignment methods for modifying the starting time or timestamps "
                "of the data if needed: "
                "https://neuroconv.readthedocs.io/en/main/user_guide/temporal_alignment.html",
                DeprecationWarning,
                stacklevel=2,
            )

        if write_as != "raw":
            warnings.warn(
                "The 'write_as' parameter is deprecated and will be removed in June 2025. "
                "NIDQ should always be written in the acquisition module of NWB. "
                "Writing data as LFP or processed data is not supported.",
                DeprecationWarning,
                stacklevel=2,
            )

        if write_electrical_series is not True:
            warnings.warn(
                "The 'write_electrical_series' parameter is deprecated and will be removed in June 2025. "
                "The option to skip the addition of the data is no longer supported. "
                "This option was used in ElectricalSeries to write the electrode and electrode group "
                "metadata without the raw data.",
                DeprecationWarning,
                stacklevel=2,
            )

        if stub_test or self.subset_channels is not None:
            recording = self.subset_recording(stub_test=stub_test)
        else:
            recording = self.recording_extractor

        if metadata is None:
            metadata = self.get_metadata()

        # Add devices
        device_metadata = metadata.get("Devices", [])
        for device in device_metadata:
            if device["name"] not in nwbfile.devices:
                nwbfile.create_device(**device)

        # Add analog and digital channels
        if self.has_analog_channels:
            self._add_analog_channels(
                nwbfile=nwbfile,
                recording=recording,
                iterator_type=iterator_type,
                iterator_opts=iterator_opts,
                always_write_timestamps=always_write_timestamps,
            )

        if self.has_digital_channels:
            self._add_digital_channels(nwbfile=nwbfile)

    def _add_analog_channels(
        self,
        nwbfile: NWBFile,
        recording,
        iterator_type: Optional[str],
        iterator_opts: Optional[dict],
        always_write_timestamps: bool,
    ):
        """
        Add analog channels from the NIDQ board to the NWB file.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to add the analog channels to
        recording : BaseRecording
            The recording extractor containing the analog channels
        iterator_type : Optional[str]
            Type of iterator to use for data streaming
        iterator_opts : Optional[dict]
            Additional options for the iterator
        always_write_timestamps : bool
            If True, always writes timestamps instead of using sampling rate
        """
        analog_recorder = recording.select_channels(channel_ids=self.analog_channel_ids)
        channel_names = analog_recorder.get_property(key="channel_names")
        segment_index = 0
        analog_data_iterator = _recording_traces_to_hdmf_iterator(
            recording=analog_recorder,
            segment_index=segment_index,
            iterator_type=iterator_type,
            iterator_opts=iterator_opts,
        )

        name = "TimeSeriesNIDQ"
        description = f"Analog data from the NIDQ board. Channels are {channel_names} in that order."
        time_series_kwargs = dict(name=name, data=analog_data_iterator, unit="a.u.", description=description)

        if always_write_timestamps:
            timestamps = recording.get_times(segment_index=segment_index)
            shifted_timestamps = timestamps
            time_series_kwargs.update(timestamps=shifted_timestamps)
        else:
            recording_has_timestamps = recording.has_time_vector(segment_index=segment_index)
            if recording_has_timestamps:
                timestamps = recording.get_times(segment_index=segment_index)
                rate = calculate_regular_series_rate(series=timestamps)
                recording_t_start = timestamps[0]
            else:
                rate = recording.get_sampling_frequency()
                recording_t_start = recording._recording_segments[segment_index].t_start or 0

            if rate:
                starting_time = float(recording_t_start)
                time_series_kwargs.update(starting_time=starting_time, rate=recording.get_sampling_frequency())
            else:
                shifted_timestamps = timestamps
                time_series_kwargs.update(timestamps=shifted_timestamps)

        time_series = TimeSeries(**time_series_kwargs)
        nwbfile.add_acquisition(time_series)

    def _add_digital_channels(self, nwbfile: NWBFile):
        """
        Add digital channels from the NIDQ board to the NWB file as events.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWB file to add the digital channels to
        """
        from ndx_events import LabeledEvents

        event_channels = self.event_extractor.channel_ids
        for channel_id in event_channels:
            events_structure = self.event_extractor.get_events(channel_id=channel_id)
            timestamps = events_structure["time"]
            labels = events_structure["label"]

            # Some channels have no events
            if timestamps.size > 0:

                # Timestamps are not ordered, the ones for off are first and then the ones for on
                ordered_indices = np.argsort(timestamps)
                ordered_timestamps = timestamps[ordered_indices]
                ordered_labels = labels[ordered_indices]

                unique_labels = np.unique(ordered_labels)
                label_to_index = {label: index for index, label in enumerate(unique_labels)}
                data = [label_to_index[label] for label in ordered_labels]

                channel_name = channel_id.split("#")[-1]
                description = f"On and Off Events from channel {channel_name}"
                name = f"EventsNIDQDigitalChannel{channel_name}"
                labeled_events = LabeledEvents(
                    name=name, description=description, timestamps=ordered_timestamps, data=data, labels=unique_labels
                )
                nwbfile.add_acquisition(labeled_events)

    def get_event_times_from_ttl(self, channel_name: str) -> np.ndarray:
        """
        Return the start of event times from the rising part of TTL pulses on one of the NIDQ channels.

        Parameters
        ----------
        channel_name : str
            Name of the channel in the .nidq.bin file.

        Returns
        -------
        rising_times : numpy.ndarray
            The times of the rising TTL pulses.
        """
        # TODO: consider RAM cost of these operations and implement safer buffering version
        rising_frames = get_rising_frames_from_ttl(
            trace=self.recording_extractor.get_traces(channel_ids=[channel_name])
        )

        nidq_timestamps = self.recording_extractor.get_times()
        rising_times = nidq_timestamps[rising_frames]

        return rising_times



from copy import deepcopy
from typing import Optional

from pydantic import DirectoryPath, validate_call
from pynwb import NWBFile

from ..basesegmentationextractorinterface import BaseSegmentationExtractorInterface
from ....utils import DeepDict


def _update_metadata_links_for_plane_segmentation_name(metadata: dict, plane_segmentation_name: str) -> DeepDict:
    """Private utility function to update the metadata with a new plane segmentation name."""
    metadata_copy = deepcopy(metadata)

    plane_segmentation_metadata = metadata_copy["Ophys"]["ImageSegmentation"]["plane_segmentations"][0]
    default_plane_segmentation_name = plane_segmentation_metadata["name"]
    default_plane_suffix = default_plane_segmentation_name.replace("PlaneSegmentation", "")
    new_plane_name_suffix = plane_segmentation_name.replace("PlaneSegmentation", "")
    imaging_plane_name = "ImagingPlane" + new_plane_name_suffix
    plane_segmentation_metadata.update(
        name=plane_segmentation_name,
        imaging_plane=imaging_plane_name,
    )
    metadata_copy["Ophys"]["ImagingPlane"][0].update(name=imaging_plane_name)

    fluorescence_metadata_per_plane = metadata_copy["Ophys"]["Fluorescence"].pop(default_plane_segmentation_name)
    # override the default name of the plane segmentation
    metadata_copy["Ophys"]["Fluorescence"][plane_segmentation_name] = fluorescence_metadata_per_plane
    trace_names = [property_name for property_name in fluorescence_metadata_per_plane.keys() if property_name != "name"]
    for trace_name in trace_names:
        default_raw_traces_name = fluorescence_metadata_per_plane[trace_name]["name"].replace(default_plane_suffix, "")
        fluorescence_metadata_per_plane[trace_name].update(name=default_raw_traces_name + new_plane_name_suffix)

    segmentation_images_metadata = metadata_copy["Ophys"]["SegmentationImages"].pop(default_plane_segmentation_name)
    metadata_copy["Ophys"]["SegmentationImages"][plane_segmentation_name] = segmentation_images_metadata
    metadata_copy["Ophys"]["SegmentationImages"][plane_segmentation_name].update(
        correlation=dict(name=f"CorrelationImage{new_plane_name_suffix}"),
        mean=dict(name=f"MeanImage{new_plane_name_suffix}"),
    )

    return metadata_copy


class Suite2pSegmentationInterface(BaseSegmentationExtractorInterface):
    """Interface for Suite2p segmentation data."""

    display_name = "Suite2p Segmentation"
    associated_suffixes = (".npy",)
    info = "Interface for Suite2p segmentation."

    @classmethod
    def get_source_schema(cls) -> dict:
        """
        Get the source schema for the Suite2p segmentation interface.

        Returns
        -------
        dict
            The schema dictionary containing input parameters and descriptions
            for initializing the Suite2p segmentation interface.
        """
        schema = super().get_source_schema()
        schema["properties"]["folder_path"][
            "description"
        ] = "Path to the folder containing Suite2p segmentation data. Should contain 'plane#' subfolder(s)."
        schema["properties"]["plane_name"][
            "description"
        ] = "The name of the plane to load. This interface only loads one plane at a time. Use the full name, e.g. 'plane0'. If this value is omitted, the first plane found will be loaded."

        return schema

    @classmethod
    def get_available_planes(cls, folder_path: DirectoryPath) -> dict:
        """
        Get the available planes in the Suite2p segmentation folder.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing Suite2p segmentation data.

        Returns
        -------
        dict
            Dictionary containing information about available planes in the dataset.
        """
        from roiextractors import Suite2pSegmentationExtractor

        return Suite2pSegmentationExtractor.get_available_planes(folder_path=folder_path)

    @classmethod
    def get_available_channels(cls, folder_path: DirectoryPath) -> dict:
        """
        Get the available channels in the Suite2p segmentation folder.

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing Suite2p segmentation data.

        Returns
        -------
        dict
            Dictionary containing information about available channels in the dataset.
        """
        from roiextractors import Suite2pSegmentationExtractor

        return Suite2pSegmentationExtractor.get_available_channels(folder_path=folder_path)

    @validate_call
    def __init__(
        self,
        folder_path: DirectoryPath,
        channel_name: Optional[str] = None,
        plane_name: Optional[str] = None,
        plane_segmentation_name: Optional[str] = None,
        verbose: bool = False,
    ):
        """

        Parameters
        ----------
        folder_path : DirectoryPath
            Path to the folder containing Suite2p segmentation data. Should contain 'plane#' sub-folders.
        channel_name: str, optional
            The name of the channel to load.
            To determine what channels are available, use ``Suite2pSegmentationInterface.get_available_channels(folder_path)``.
        plane_name: str, optional
            The name of the plane to load. This interface only loads one plane at a time.
            If this value is omitted, the first plane found will be loaded.
            To determine what planes are available, use ``Suite2pSegmentationInterface.get_available_planes(folder_path)``.
        plane_segmentation_name: str, optional
            The name of the plane segmentation to be added.
        """

        super().__init__(folder_path=folder_path, channel_name=channel_name, plane_name=plane_name)
        available_planes = self.get_available_planes(folder_path=self.source_data["folder_path"])
        available_channels = self.get_available_channels(folder_path=self.source_data["folder_path"])

        if plane_segmentation_name is None:
            plane_segmentation_name = (
                "PlaneSegmentation"
                if len(available_planes) == 1 and len(available_channels) == 1
                else f"PlaneSegmentation{self.segmentation_extractor.channel_name.capitalize()}{self.segmentation_extractor.plane_name.capitalize()}"
            )

        self.plane_segmentation_name = plane_segmentation_name
        self.verbose = verbose

    def get_metadata(self) -> DeepDict:
        """
        Get metadata for the Suite2p segmentation data.

        Returns
        -------
        DeepDict
            Dictionary containing metadata including plane segmentation details,
            fluorescence data, and segmentation images.
        """
        metadata = super().get_metadata()

        # No need to update the metadata links for the default plane segmentation name
        default_plane_segmentation_name = metadata["Ophys"]["ImageSegmentation"]["plane_segmentations"][0]["name"]
        if self.plane_segmentation_name == default_plane_segmentation_name:
            return metadata

        metadata = _update_metadata_links_for_plane_segmentation_name(
            metadata=metadata,
            plane_segmentation_name=self.plane_segmentation_name,
        )

        return metadata

    def add_to_nwbfile(
        self,
        nwbfile: NWBFile,
        metadata: Optional[dict] = None,
        stub_test: bool = False,
        stub_frames: int = 100,
        include_roi_centroids: bool = True,
        include_roi_acceptance: bool = True,
        mask_type: Optional[str] = "image",  # Literal["image", "pixel", "voxel"]
        plane_segmentation_name: Optional[str] = None,
        iterator_options: Optional[dict] = None,
    ):
        """
        Add segmentation data to the specified NWBFile.

        Parameters
        ----------
        nwbfile : NWBFile
            The NWBFile object to which the segmentation data will be added.
        metadata : dict, optional
            Metadata containing information about the segmentation. If None, default metadata is used.
        stub_test : bool, optional
            If True, only a subset of the data (defined by `stub_frames`) will be added for testing purposes,
            by default False.
        stub_frames : int, optional
            The number of frames to include in the subset if `stub_test` is True, by default 100.
        include_roi_centroids : bool, optional
            Whether to include the centroids of regions of interest (ROIs) in the data, by default True.
        include_roi_acceptance : bool, optional
            Whether to include acceptance status of ROIs, by default True.
        mask_type : str, default: 'image'
            There are three types of ROI masks in NWB, 'image', 'pixel', and 'voxel'.

            * 'image' masks have the same shape as the reference images the segmentation was applied to, and weight each pixel
            by its contribution to the ROI (typically boolean, with 0 meaning 'not in the ROI').
            * 'pixel' masks are instead indexed by ROI, with the data at each index being the shape of the image by the number
            of pixels in each ROI.
            * 'voxel' masks are instead indexed by ROI, with the data at each index being the shape of the volume by the number
            of voxels in each ROI.

            Specify your choice between these two as mask_type='image', 'pixel', 'voxel', or None.
            plane_segmentation_name : str, optional
            The name of the plane segmentation object, by default None.
        iterator_options : dict, optional
            Additional options for iterating over the data, by default None.
        """
        super().add_to_nwbfile(
            nwbfile=nwbfile,
            metadata=metadata,
            stub_test=stub_test,
            stub_frames=stub_frames,
            include_roi_centroids=include_roi_centroids,
            include_roi_acceptance=include_roi_acceptance,
            mask_type=mask_type,
            plane_segmentation_name=self.plane_segmentation_name,
            iterator_options=iterator_options,
        )
